#include <WiFi.h>
#include <WebSocketsServer.h>

#define BUFFERSIZE 512

// Set your Access Point network credentials
const char *ssid = "EcoTechNode";
const char *password = "123456789";

const int PinA = 26;
const int PinB = 27;
const int PinWatchDog = 25;

// Create a WebSocket server object
WebSocketsServer webSocket = WebSocketsServer(81);

byte modbusAddress = 0x01;  // The sensor's modbus address, or SlaveID
long modbusBaudRate = 9600; // The baud rate the sensor uses

struct portConfig
{
  long baudRate;
  byte numBits;
  byte parity;
  byte stopBits;
  bool update;
};

int activePort = 0;
uint8_t txBuffer[BUFFERSIZE];
int txBufferLength = 0;
uint8_t rxBuffer[BUFFERSIZE];
int rxBufferLength = 0;
bool bufferReady = false;
bool bufferWaitReply = false;

portConfig portA = {9600, 8, 0, 1, true};
portConfig portB = {9600, 8, 0, 1, true};

void setup()
{
  pinMode(PinA, OUTPUT);
  pinMode(PinB, OUTPUT);
  pinMode(PinWatchDog, OUTPUT);
  digitalWrite(PinA, LOW);
  digitalWrite(PinB, LOW);
  digitalWrite(PinWatchDog, LOW);

  Serial.begin(115200);
  Serial1.begin(portA.baudRate, parseSerialConfig(portA));
  Serial2.begin(portB.baudRate, parseSerialConfig(portB));

  //  modbus.setDebugStream(&Serial);
  //  modbus.begin(modbusAddress, modbusSerial, DEREPin);

  // Setup ESP32 as an Access Point
  String ssid = createUniqueHotspotName();
  WiFi.softAP(ssid, password);
  Serial.print("AP Started ");
  Serial.println(ssid);
  Serial.print("IP Address: ");
  Serial.println(WiFi.softAPIP());

  // Start WebSocket server and assign callback
  webSocket.begin();
  webSocket.onEvent(webSocketEvent);
}

unsigned long myTime;
unsigned long hotdogTimer;
unsigned long hotdogOffTimer;
bool hotdogState = false;

void loop()
{
  // Look for and handle WebSocket data
  webSocket.loop();

  TxRoutine();
  RxRoutine();

  if ((millis() - hotdogTimer >= 500) && !hotdogState)
  {
    hotdogTimer = millis();
    hotdogOffTimer = millis();
    hotdogState = true;
    digitalWrite(PinWatchDog, hotdogState);
  }
  if ((millis() - hotdogOffTimer >= 100) && hotdogState)
  {
    hotdogState = false;
    digitalWrite(PinWatchDog, hotdogState);
  }
}

// Callback: when receiving any WebSocket message
void webSocketEvent(uint8_t num, WStype_t type, uint8_t *payload, size_t length)
{
  // Figure out the type of WebSocket event
  switch (type)
  {
    case WStype_DISCONNECTED:
      Serial.printf("[%u] Disconnected!\n", num);
      break;
    case WStype_CONNECTED:
      {
        IPAddress ip = webSocket.remoteIP(num);
        Serial.printf("[%u] Connection from ", num);
        Serial.println(ip.toString());
      }
      break;
    case WStype_TEXT:
      parseSocketText(payload, length);
      Serial.printf("[%u] Text: %s\n", num, payload);
      break;
    case WStype_BIN:
      parseSocketBinary(payload, length);
      break;
    case WStype_ERROR:
      Serial.println("Socket error");
      break;
    case WStype_FRAGMENT_TEXT_START:
      break;
    case WStype_FRAGMENT_BIN_START:
      break;
    case WStype_FRAGMENT:
      break;
    case WStype_FRAGMENT_FIN:
      break;
    case WStype_PING:
      break;
    case WStype_PONG:
      break;
  }
}

void parseSocketText(uint8_t *payload, size_t length)
{
  // Parse text as cmds
  // 1. Set port A config
  // msg like "A,9600,8,0,1"
  // 2. Set port B config
  // msg like "B,9600,8,0,1"
  // 3. Set active port
  // msg like "P,A"

  if (strstr(payload, "A"))
  {
    sscanf(payload, "A,%ld,%d,%d,%d", &portA.baudRate, &portA.numBits, &portA.parity, &portA.stopBits);
    portA.update = true;
  }
  else if (strstr(payload, "B"))
  {
    sscanf(payload, "B,%ld,%d,%d,%d", &portB.baudRate, &portB.numBits, &portB.parity, &portB.stopBits);
    portB.update = true;
  }
  else if (strstr(payload, "P"))
  {
    activePort = (strstr(payload, "A")) ? 0 : 1;
  }
  else
  {
    Serial.println("Unknown command");
  }
}

void parseSocketBinary(uint8_t *payload, size_t length)
{
  if (length > BUFFERSIZE)
  {
    Serial.println("Buffer overflow");
    return;
  }
  if (length < 3)
  {
    Serial.println("Invalid message");
    return;
  }
  memcpy(&txBuffer, &payload, length);
  txBufferLength = length;
  bufferReady = true;
}

void TxRoutine()
{
  if (bufferReady)
  {
    if (activePort == 0)
    {
      sendData(Serial1, txBuffer, PinA, BUFFERSIZE);
    }
    else if (activePort == 1)
    {
      sendData(Serial2, txBuffer, PinB, BUFFERSIZE);
    }
    bufferReady = false;
    bufferWaitReply = true;
  }
}
void RxRoutine()
{
  if (bufferWaitReply)
  {
    if (activePort == 0)
    {
      rxBufferLength = receiveData(Serial1, rxBuffer, BUFFERSIZE);
      webSocket.broadcastBIN(rxBuffer, rxBufferLength);
    }
    else if (activePort == 1)
    {
      rxBufferLength = receiveData(Serial2, rxBuffer, BUFFERSIZE);
      webSocket.broadcastBIN(rxBuffer, rxBufferLength);
    }
    bufferWaitReply = false;
  }
}

void sendData(HardwareSerial &serial, byte *buffer, int controlPin, int bufferSize)
{
  // Enable transmitter
  digitalWrite(controlPin, HIGH);

  for (int i = 0; i < bufferSize; i++)
  {
    if (buffer[i] != 0)
    { // Assuming 0 is not a valid data byte; adjust as needed
      serial.write(buffer[i]);
    }
  }

  // Ensure data is transmitted before disabling transmitter
  serial.flush();
  delay(2);
  // Disable transmitter, enable receiver
  digitalWrite(controlPin, LOW);
}
bool receiveData(HardwareSerial &serial, byte *buffer, int bufferSize)
{
  unsigned long startTime = millis();
  int index = 0;
  const unsigned long timeout = 500;

  while ((millis() - startTime) < timeout && index < bufferSize)
  {
    if (serial.available())
    {
      buffer[index++] = serial.read();
      startTime = millis();
    }
  }

  return (index > 0);
}
SerialConfig parseSerialConfig(const portConfig &config)
{
  switch (config.numBits)
  {
    case 5:
      switch (config.parity)
      {
        case 0:
          return config.stopBits == 1 ? SERIAL_5N1 : SERIAL_5N2;
        case 1:
          return config.stopBits == 1 ? SERIAL_5E1 : SERIAL_5E2;
        case 2:
          return config.stopBits == 1 ? SERIAL_5O1 : SERIAL_5O2;
      }
      break;
    case 6:
      switch (config.parity)
      {
        case 0:
          return config.stopBits == 1 ? SERIAL_6N1 : SERIAL_6N2;
        case 1:
          return config.stopBits == 1 ? SERIAL_6E1 : SERIAL_6E2;
        case 2:
          return config.stopBits == 1 ? SERIAL_6O1 : SERIAL_6O2;
      }
      break;
    case 7:
      switch (config.parity)
      {
        case 0:
          return config.stopBits == 1 ? SERIAL_7N1 : SERIAL_7N2;
        case 1:
          return config.stopBits == 1 ? SERIAL_7E1 : SERIAL_7E2;
        case 2:
          return config.stopBits == 1 ? SERIAL_7O1 : SERIAL_7O2;
      }
      break;
    case 8:
      switch (config.parity)
      {
        case 0:
          return config.stopBits == 1 ? SERIAL_8N1 : SERIAL_8N2;
        case 1:
          return config.stopBits == 1 ? SERIAL_8E1 : SERIAL_8E2;
        case 2:
          return config.stopBits == 1 ? SERIAL_8O1 : SERIAL_8O2;
      }
      break;
  }

  return SERIAL_8N1; // Default case
}
String createUniqueHotspotName()
{
  // Get the MAC address
  byte mac[6];
  WiFi.macAddress(mac);

  // Convert MAC address to a String
  String macID = "";
  for (int i = 0; i < 6; i++)
  {
    if (mac[i] < 16)
    {
      macID += "0"; // Add leading zero for single digit hex values
    }
    macID += String(mac[i], HEX);
  }

  // Create SSID by appending a part of the MAC address to the base name
  String ssid = "EcoTech_" + macID.substring(6); // Using the last part of the MAC address
  return ssid;
}
