package fr.drangies.cordova.serial;

import java.io.DataOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.cordova.CallbackContext;
import org.apache.cordova.CordovaPlugin;
import org.apache.cordova.PluginResult;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.hoho.android.usbserial.driver.CdcAcmSerialDriver;
import com.hoho.android.usbserial.driver.Ch34xSerialDriver;
import com.hoho.android.usbserial.driver.Cp21xxSerialDriver;
import com.hoho.android.usbserial.driver.FtdiSerialDriver;
import com.hoho.android.usbserial.driver.ProbeTable;
import com.hoho.android.usbserial.driver.ProlificSerialDriver;
import com.hoho.android.usbserial.driver.UsbSerialDriver;
import com.hoho.android.usbserial.driver.UsbSerialPort;
import com.hoho.android.usbserial.driver.UsbSerialProber;
import com.hoho.android.usbserial.util.SerialInputOutputManager;

import android.R.integer;
import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbDeviceConnection;
import android.hardware.usb.UsbManager;
import android.util.Base64;
import android.util.Log;

/**
 * Cordova plugin to communicate with the android serial port
 * 
 * @author Xavier Seignard <xavier.seignard@gmail.com>
 */
public class Serial extends CordovaPlugin {
	// logging tag
	private final String TAG = Serial.class.getSimpleName();
	// actions definitions
	private static final String ACTION_REQUEST_PERMISSION = "requestPermission";
	private static final String ACTION_OPEN = "openSerial";
	private static final String ACTION_READ = "readSerial";
	private static final String ACTION_WRITE = "writeSerial";
	private static final String ACTION_WRITE_HEX = "writeSerialHex";
	private static final String ACTION_CLOSE = "closeSerial";
	private static final String ACTION_READ_CALLBACK = "registerReadCallback";

	// UsbManager instance to deal with permission and opening
	private UsbManager manager;
	// The current driver that handle the serial port
	private UsbSerialDriver driver;
	// The serial port that will be used in this plugin
	// private UsbSerialPort port;
	private UsbSerialPort portA;
	private UsbSerialPort portB;
	// Read buffer, and read params
	private static final int READ_WAIT_MILLIS = 1000;
	private static final int BUFSIZ = 4096;
	private final ByteBuffer mReadBufferA = ByteBuffer.allocate(BUFSIZ);
	private final ByteBuffer mReadBufferB = ByteBuffer.allocate(BUFSIZ);
	// Connection info
	private int baudRateA;
	private int dataBitsA;
	private int stopBitsA;
	private int parityA;
	private boolean setDTRA;
	private boolean setRTSA;
	private int baudRateB;
	private int dataBitsB;
	private int stopBitsB;
	private int parityB;
	private boolean setDTRB;
	private boolean setRTSB;
	private boolean sleepOnPause;
	private boolean portAOpen = false;
	private boolean portBOpen = false;

	// callback that will be used to send back data to the cordova app
	private CallbackContext readCallback;

	// I/O manager to handle new incoming serial data
	private final ExecutorService mExecutorA = Executors.newSingleThreadExecutor();
	private final ExecutorService mExecutorB = Executors.newSingleThreadExecutor();
	private SerialInputOutputManager mSerialIoManagerA;
	private SerialInputOutputManager mSerialIoManagerB;
	private final SerialInputOutputManager.Listener mListenerA = new SerialInputOutputManager.Listener() {
		@Override
		public void onRunError(Exception e) {
			Log.d(TAG, "Runner A stopped.");
		}

		@Override
		public void onNewData(final byte[] data) {
			// Handle data for port A
			Serial.this.updateReceivedDataA(data);
		}
	};

	private final SerialInputOutputManager.Listener mListenerB = new SerialInputOutputManager.Listener() {
		@Override
		public void onRunError(Exception e) {
			Log.d(TAG, "Runner B stopped.");
		}

		@Override
		public void onNewData(final byte[] data) {
			// Handle data for port B
			Serial.this.updateReceivedDataB(data);
		}
	};

	/**
	 * Overridden execute method
	 * 
	 * @param action          the string representation of the action to execute
	 * @param args
	 * @param callbackContext the cordova {@link CallbackContext}
	 * @return true if the action exists, false otherwise
	 * @throws JSONException if the args parsing fails
	 */
	@Override
	public boolean execute(String action, JSONArray args, final CallbackContext callbackContext) throws JSONException {
		Log.d(TAG, "mAction: " + action);
		JSONObject arg_object = args.optJSONObject(0);
		Log.d(TAG, args.toString());
		// request permission
		Integer dev = arg_object.has("dev") ? arg_object.getInt("dev") : null;
		Log.d(TAG, "dev: " + dev);
		if (ACTION_REQUEST_PERMISSION.equals(action)) {
			JSONObject opts = arg_object.has("opts") ? arg_object.getJSONObject("opts") : new JSONObject();
			requestPermission(opts, callbackContext);
			return true;
		}
		// open serial port
		else if (ACTION_OPEN.equals(action)) {
			JSONObject opts = arg_object.has("opts") ? arg_object.getJSONObject("opts") : new JSONObject();
			openSerial(dev, opts, callbackContext);
			return true;
		}
		// write to the serial port
		else if (ACTION_WRITE.equals(action)) {
			String data = arg_object.getString("data");
			Log.d(TAG, "got data: " + data);
			writeSerial(dev, data, callbackContext);
			return true;
		}
		// write hex to the serial port
		else if (ACTION_WRITE_HEX.equals(action)) {
			String data = arg_object.getString("data");
			writeSerialHex(dev, data, callbackContext);
			return true;
		}
		// read on the serial port
		else if (ACTION_READ.equals(action)) {
			readSerial(dev, callbackContext);
			return true;
		}
		// close the serial port
		else if (ACTION_CLOSE.equals(action)) {
			closeSerial(dev, callbackContext);
			return true;
		}
		// Register read callback
		else if (ACTION_READ_CALLBACK.equals(action)) {
			registerReadCallback(callbackContext);
			return true;
		}
		// the action doesn't exist
		return false;
	}

	/**
	 * Request permission the the user for the app to use the USB/serial port
	 * 
	 * @param callbackContext the cordova {@link CallbackContext}
	 */
	private void requestPermission(final JSONObject opts, final CallbackContext callbackContext) {
		cordova.getThreadPool().execute(new Runnable() {
			public void run() {
				changeHostMode();
				// get UsbManager from Android
				manager = (UsbManager) cordova.getActivity().getSystemService(Context.USB_SERVICE);
				UsbSerialProber prober;

				if (opts.has("vid") && opts.has("pid")) {
					ProbeTable customTable = new ProbeTable();
					Object o_vid = opts.opt("vid"); // can be an integer Number or a hex String
					Object o_pid = opts.opt("pid"); // can be an integer Number or a hex String
					int vid = o_vid instanceof Number ? ((Number) o_vid).intValue()
							: Integer.parseInt((String) o_vid, 16);
					int pid = o_pid instanceof Number ? ((Number) o_pid).intValue()
							: Integer.parseInt((String) o_pid, 16);
					String driver = opts.has("driver") ? (String) opts.opt("driver") : "FtdiSerialDriver";// "CdcAcmSerialDriver";

					if (driver.equals("FtdiSerialDriver")) {
						customTable.addProduct(vid, pid, FtdiSerialDriver.class);
					} else if (driver.equals("CdcAcmSerialDriver")) {
						customTable.addProduct(vid, pid, CdcAcmSerialDriver.class);
					} else if (driver.equals("Cp21xxSerialDriver")) {
						customTable.addProduct(vid, pid, Cp21xxSerialDriver.class);
					} else if (driver.equals("ProlificSerialDriver")) {
						customTable.addProduct(vid, pid, ProlificSerialDriver.class);
					} else if (driver.equals("Ch34xSerialDriver")) {
						customTable.addProduct(vid, pid, Ch34xSerialDriver.class);
					} else {
						Log.d(TAG, "Unknown driver!");
						callbackContext.error("Unknown driver!");
					}

					prober = new UsbSerialProber(customTable);

				} else {
					// find all available drivers from attached devices.
					prober = UsbSerialProber.getDefaultProber();
				}

				List<UsbSerialDriver> availableDrivers = prober.findAllDrivers(manager);


				// if (!availableDrivers.isEmpty()) {
				// 	// get the first one as there is a high chance that there is no more than one
				// 	// usb device attached to your android
				// 	Log.d(TAG, "Device found!");
				// 	driver = availableDrivers.get(0);
				// 	UsbDevice device = driver.getDevice();
				// 	// create the intent that will be used to get the permission
				// 	// PendingIntent pendingIntent =
				// 	// PendingIntent.getBroadcast(cordova.getActivity(), 0,
				// 	// new Intent(UsbBroadcastReceiver.USB_PERMISSION), 0);
				// 	PendingIntent pendingIntent = PendingIntent.getBroadcast(cordova.getActivity(), 0,
				// 			new Intent(UsbBroadcastReceiver.USB_PERMISSION), PendingIntent.FLAG_UPDATE_CURRENT);
				// 	// and a filter on the permission we ask
				// 	IntentFilter filter = new IntentFilter();
				// 	filter.addAction(UsbBroadcastReceiver.USB_PERMISSION);
				// 	// this broadcast receiver will handle the permission results
				// 	UsbBroadcastReceiver usbReceiver = new UsbBroadcastReceiver(callbackContext, cordova.getActivity());
				// 	cordova.getActivity().registerReceiver(usbReceiver, filter);
				// 	// finally ask for the permission
				// 	manager.requestPermission(device, pendingIntent);
				// } else {
				// 	// no available drivers
				// 	Log.d(TAG, "No device found!");
				// 	callbackContext.error("No device found!");
				// }
				if (!availableDrivers.isEmpty()) {
					Log.d(TAG, "Device found!");
					driver = availableDrivers.get(0);
					UsbDevice device = driver.getDevice();
	
					if (manager.hasPermission(device)) {
						// Already have permission, so continue with your operation.
						// For instance, you might notify the callbackContext that everything is okay.
						callbackContext.success("Device permission already granted.");
					} else {
						// Fallback to manually request permission (although with the manifest filter, this scenario is less likely).
						PendingIntent pendingIntent = PendingIntent.getBroadcast(cordova.getActivity(), 0,
								new Intent(UsbBroadcastReceiver.USB_PERMISSION), PendingIntent.FLAG_MUTABLE);
						IntentFilter filter = new IntentFilter();
						filter.addAction(UsbBroadcastReceiver.USB_PERMISSION);
						UsbBroadcastReceiver usbReceiver = new UsbBroadcastReceiver(callbackContext, cordova.getActivity());
						cordova.getActivity().registerReceiver(usbReceiver, filter);
						manager.requestPermission(device, pendingIntent);
					}
				} else {
					Log.d(TAG, "No device found!");
					callbackContext.error("No device found!");
				}
			}
		});
	}

	private void changeHostMode() {
		try {
			Process su = Runtime.getRuntime().exec("su");
			DataOutputStream outputStream = new DataOutputStream(su.getOutputStream());
			outputStream.writeBytes("echo dfp > /sys/class/dual_role_usb/dual-role-type_c_port0/mode\n");
			outputStream.flush();
			outputStream.writeBytes("echo host > /sys/class/dual_role_usb/dual-role-type_c_port0/data_role\n");
			outputStream.flush();
			outputStream.close();
			su.waitFor();
		} catch (IOException e) {
			// deal with error
			Log.d(TAG, e.getMessage());
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		try {
			Process su = Runtime.getRuntime().exec("su");
			DataOutputStream outputStream = new DataOutputStream(su.getOutputStream());
			outputStream.writeBytes("busybox echo dfp > /sys/class/dual_role_usb/dual-role-type_c_port0/mode\n");
			outputStream.flush();
			outputStream.writeBytes("busybox echo host > /sys/class/dual_role_usb/dual-role-type_c_port0/data_role\n");
			outputStream.flush();
			outputStream.close();
			su.waitFor();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		try {
			// Execute the command to change mode
			Process process1 = Runtime.getRuntime()
					.exec("echo dfp > /sys/class/dual_role_usb/dual-role-type_c_port0/mode");
			process1.waitFor();

			// Execute the command to change data role
			Process process2 = Runtime.getRuntime()
					.exec("echo host > /sys/class/dual_role_usb/dual-role-type_c_port0/data_role");
			process2.waitFor();

		} catch (IOException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Open the serial port from Cordova
	 * 
	 * @param opts            a {@link JSONObject} containing the connection
	 *                        paramters
	 * @param callbackContext the cordova {@link CallbackContext}
	 */
	private void openSerial(final Integer deviceNumber, final JSONObject opts, final CallbackContext callbackContext) {
		cordova.getThreadPool().execute(new Runnable() {
			public void run() {
				UsbDeviceConnection connection = manager.openDevice(driver.getDevice());
				if (connection != null) {
					// get first port and open it
					Log.d(TAG, "Device " + deviceNumber + " connect!");
					switch (deviceNumber) {
						case 0:
							portA = driver.getPorts().get(0);
							try {
								// get connection params or the default values
								baudRateA = opts.has("baudRate") ? opts.getInt("baudRate") : 9600;
								dataBitsA = opts.has("dataBits") ? opts.getInt("dataBits") : UsbSerialPort.DATABITS_8;
								stopBitsA = opts.has("stopBits") ? opts.getInt("stopBits") : UsbSerialPort.STOPBITS_1;
								parityA = opts.has("parity") ? opts.getInt("parity") : UsbSerialPort.PARITY_NONE;
								setDTRA = false; // opts.has("dtr") && opts.getBoolean("dtr");
								setRTSA = false; // opts.has("rts") && opts.getBoolean("rts");
								// Sleep On Pause defaults to true
								sleepOnPause = opts.has("sleepOnPause") ? opts.getBoolean("sleepOnPause") : false;

								portA.open(connection);
								portA.setParameters(baudRateA, dataBitsA, stopBitsA, parityA);
								Log.d(TAG, "Open port A with baudRate: " + baudRateA + ", dataBits: " + dataBitsA
										+ ", stopBits: " + stopBitsA + ", parity: " + parityA + ", setDTR: " + setDTRA
										+ ", setRTS: " + setRTSA);
								if (setDTRA)
									portA.setDTR(true);
								if (setRTSA)
									portA.setRTS(true);
							} catch (IOException e) {
								// deal with basic error
								Log.d(TAG, e.getMessage());
								callbackContext.error(e.getMessage());
							} catch (JSONException e) {
								// deal with json error
								Log.d(TAG, e.getMessage());
								callbackContext.error(e.getMessage());
							}
							break;
						case 1:
							portB = driver.getPorts().get(1);
							try {
								// get connection params or the default values
								baudRateB = opts.has("baudRate") ? opts.getInt("baudRate") : 9600;
								dataBitsB = opts.has("dataBits") ? opts.getInt("dataBits") : UsbSerialPort.DATABITS_8;
								stopBitsB = opts.has("stopBits") ? opts.getInt("stopBits") : UsbSerialPort.STOPBITS_1;
								parityB = opts.has("parity") ? opts.getInt("parity") : UsbSerialPort.PARITY_NONE;
								setDTRB = false;// opts.has("dtr") && opts.getBoolean("dtr");
								setRTSB = false;// opts.has("rts") && opts.getBoolean("rts");
								// Sleep On Pause defaults to true
								sleepOnPause = opts.has("sleepOnPause") ? opts.getBoolean("sleepOnPause") : false;

								portB.open(connection);
								portB.setParameters(baudRateB, dataBitsB, stopBitsB, parityB);
								Log.d(TAG, "Open port B with baudRate: " + baudRateB + ", dataBits: " + dataBitsB
										+ ", stopBits: " + stopBitsB + ", parity: " + parityB + ", setDTR: " + setDTRB
										+ ", setRTS: " + setRTSB);
								if (setDTRB)
									portB.setDTR(true);
								if (setRTSB)
									portB.setRTS(true);
							} catch (IOException e) {
								// deal with basic error
								Log.d(TAG, e.getMessage());
								callbackContext.error(e.getMessage());
							} catch (JSONException e) {
								// deal with json error
								Log.d(TAG, e.getMessage());
								callbackContext.error(e.getMessage());
							}
							break;
						default:
							Log.e(TAG, "Invalid device number provided.");
							return; // or handle this case as you see fit.
					}

					Log.d(TAG, "Serial port " + deviceNumber + " opened!");
					callbackContext.success("Serial port opened!");
				} else {
					Log.d(TAG, "Cannot connect to the device!");
					callbackContext.error("Cannot connect to the device!");
				}
				onDeviceStateChange();
			}
		});
	}

	/**
	 * Write on the serial port
	 * 
	 * @param data            the {@link String} representation of the data to be
	 *                        written on the port
	 * @param callbackContext the cordova {@link CallbackContext}
	 */
	private void writeSerial(final Integer dev, final String data, final CallbackContext callbackContext) {
		cordova.getThreadPool().execute(new Runnable() {
			public void run() {
				if (dev == 0) {
					if (portA == null) {
						callbackContext.error("Writing a closed port.");
					} else {
						try {
							byte[] buffer = data.getBytes();
							Log.d(TAG, "Writing to port A: " + data);
							portA.write(buffer, 1000);

							callbackContext.success();
						} catch (IOException e) {
							// deal with error
							Log.d(TAG, e.getMessage());
							callbackContext.error(e.getMessage());
						}
					}
				}
				if (dev == 1) {
					if (portB == null) {
						callbackContext.error("Writing a closed port.");
					} else {
						try {
							byte[] buffer = data.getBytes();
							Log.d(TAG, "Writing to port B: " + data);
							portB.write(buffer, 1000);

							callbackContext.success();
						} catch (IOException e) {
							// deal with error
							Log.d(TAG, e.getMessage());
							callbackContext.error(e.getMessage());
						}
					}
				}
			}
		});
	}

	/**
	 * Write hex on the serial port
	 * 
	 * @param data            the {@link String} representation of the data to be
	 *                        written on the port as hexadecimal string
	 *                        e.g. "ff55aaeeef000233"
	 * @param callbackContext the cordova {@link CallbackContext}
	 */
	private void writeSerialHex(final Integer dev, final String data, final CallbackContext callbackContext) {
		cordova.getThreadPool().execute(new Runnable() {
			public void run() {
				if (dev == 0) {
					if (portA == null) {
						callbackContext.error("Writing a closed port.");
					} else {
						try {
							Log.d(TAG, "A: " + data);
							byte[] buffer = hexStringToByteArray(data);
							portA.write(buffer, 1000);
							callbackContext.success(" bytes written.");
						} catch (IOException e) {
							// deal with error
							Log.d(TAG, e.getMessage());
							callbackContext.error(e.getMessage());
						}
					}
				}
				if (dev == 1) {
					if (portB == null) {
						callbackContext.error("Writing a closed port.");
					} else {
						try {
							Log.d(TAG, "B: " + data);
							byte[] buffer = hexStringToByteArray(data);
							portB.write(buffer, 1000);
							callbackContext.success(" bytes written.");
						} catch (IOException e) {
							// deal with error
							Log.d(TAG, e.getMessage());
							callbackContext.error(e.getMessage());
						}
					}
				}
			}

		});
	}

	/**
	 * Convert a given string of hexadecimal numbers
	 * into a byte[] array where every 2 hex chars get packed into
	 * a single byte.
	 *
	 * E.g. "ffaa55" results in a 3 byte long byte array
	 *
	 * @param s
	 * @return
	 */
	private byte[] hexStringToByteArray(String s) {
		int len = s.length();
		byte[] data = new byte[len / 2];
		for (int i = 0; i < len; i += 2) {
			data[i / 2] = (byte) ((Character.digit(s.charAt(i), 16) << 4)
					+ Character.digit(s.charAt(i + 1), 16));
		}
		return data;
	}

	/**
	 * Read on the serial port
	 * 
	 * @param callbackContext the {@link CallbackContext}
	 */
	private void readSerial(final Integer deviceId, final CallbackContext callbackContext) {
		cordova.getThreadPool().execute(new Runnable() {
			public void run() {
				if (deviceId == 0) {
					if (portA == null) {
						callbackContext.error("Reading a closed port.");
					} else {
						try {
							int len = 0;
							if (deviceId == 0) {
								len = portA.read(mReadBufferA.array(), READ_WAIT_MILLIS);
							}
							// Whatever happens, we send an "OK" result, up to the
							// receiver to check that len > 0
							PluginResult.Status status = PluginResult.Status.OK;
							byte[] data;
							if (len > 0) {
								Log.d(TAG, "Read data len=" + len);
								data = new byte[len];
								mReadBufferA.get(data, 0, len);
								mReadBufferA.clear();
							} 
							else {
								data = new byte[0];
							}
							JSONArray dataArray = new JSONArray();
							for (byte b : data) {
								dataArray.put((int) b & 0xFF); // Convert byte to unsigned int before adding
							}
							JSONObject resp = new JSONObject();
							resp.put("data", dataArray);
							resp.put("port", "0");

							callbackContext.sendPluginResult(new PluginResult(status, resp));
						} catch (IOException e) {
							// deal with error
							Log.d(TAG, e.getMessage());
							callbackContext.error(e.getMessage());
						} catch (JSONException e) {
							Log.e(TAG, "JSON error: " + e.getMessage());
							callbackContext.error("JSON error: " + e.getMessage());
						}
					}
				}
				if (deviceId == 1) {
					Log.d(TAG, "Reading from port B");
					if (portB == null) {
						callbackContext.error("Reading a closed port.");
					} else {
						try {
							int len = 0;
							if (deviceId == 1) {
								len = portB.read(mReadBufferB.array(), READ_WAIT_MILLIS);
							}
							// Whatever happens, we send an "OK" result, up to the
							// receiver to check that len > 0
							PluginResult.Status status = PluginResult.Status.OK;
							byte[] data;
							if (len > 0) {
								Log.d(TAG, "Read data len=" + len);
								data = new byte[len];
								mReadBufferB.get(data, 0, len);
								mReadBufferB.clear();
							} 
							else {
								data = new byte[0];
							}
							JSONArray dataArray = new JSONArray();
							for (byte b : data) {
								dataArray.put((int) b & 0xFF); // Convert byte to unsigned int before adding
							}
							JSONObject resp = new JSONObject();
							resp.put("data", dataArray);
							resp.put("port", "1");

							callbackContext.sendPluginResult(new PluginResult(status, resp));
						} catch (IOException e) {
							// deal with error
							Log.d(TAG, e.getMessage());
							callbackContext.error(e.getMessage());
						} catch (JSONException e) {
							Log.e(TAG, "JSON error: " + e.getMessage());
							callbackContext.error("JSON error: " + e.getMessage());
						}
					}
				}
			}
		});
	}

	/**
	 * Close the serial port
	 * 
	 * @param callbackContext the cordova {@link CallbackContext}
	 */
	private void closeSerial(final Integer deviceId, final CallbackContext callbackContext) {
		cordova.getThreadPool().execute(new Runnable() {
			public void run() {
				if (deviceId == 0) {
					try {
						// Make sure we don't die if we try to close an non-existing port!
						if (portA != null) {
							portA.setDTR(false);
							portA.setRTS(false);
							portA.close();
						}
						// port = null;
						callbackContext.success();
					} catch (IOException e) {
						// deal with error
						Log.d(TAG, e.getMessage());
						callbackContext.error(e.getMessage());
					}
				}
				if (deviceId == 1) {
					try {
						// Make sure we don't die if we try to close an non-existing port!
						if (portB != null) {
							portB.setDTR(false);
							portB.setRTS(false);
							portB.close();
						}
						// port = null;
						callbackContext.success();
					} catch (IOException e) {
						// deal with error
						Log.d(TAG, e.getMessage());
						callbackContext.error(e.getMessage());
					}
				}
				onDeviceStateChange();
			}
		});
	}

	/**
	 * Stop observing serial connection
	 */
	private void stopIoManager() {
		if (mSerialIoManagerA != null) {
			Log.i(TAG, "Stopping io manager A.");
			mSerialIoManagerA.stop();
			mSerialIoManagerA = null;
		}
		if (mSerialIoManagerB != null) {
			Log.i(TAG, "Stopping io manager A.");
			mSerialIoManagerB.stop();
			mSerialIoManagerB = null;
		}
	}

	/**
	 * Observe serial connection
	 */
	private void startIoManager() {
		if (driver != null) {
			Log.i(TAG, "Starting io managers");
			if (portAOpen) {
				Log.i(TAG, "Starting io manager A.");
				mSerialIoManagerA = new SerialInputOutputManager(portA, mListenerA);
				mExecutorA.submit(mSerialIoManagerA);
			}
			if (portBOpen) {
				Log.i(TAG, "Starting io manager B.");
				mSerialIoManagerB = new SerialInputOutputManager(portB, mListenerB);
				mExecutorB.submit(mSerialIoManagerB);
			}
		}
	}

	/**
	 * Restart the observation of the serial connection
	 */
	private void onDeviceStateChange() {
		stopIoManager();
		startIoManager();
	}

	/**
	 * Dispatch read data to javascript
	 * 
	 * @param data the array of bytes to dispatch
	 */
	private void updateReceivedDataA(byte[] data) {
		if (readCallback != null) {
			PluginResult result = new PluginResult(PluginResult.Status.OK, data);
			result.setKeepCallback(true);
			readCallback.sendPluginResult(result);
		}
	}

	private void updateReceivedDataB(byte[] data) {
		if (readCallback != null) {
			PluginResult result = new PluginResult(PluginResult.Status.OK, data);
			result.setKeepCallback(true);
			readCallback.sendPluginResult(result);
		}
	}

	/**
	 * Register callback for read data
	 * 
	 * @param callbackContext the cordova {@link CallbackContext}
	 */
	private void registerReadCallback(final CallbackContext callbackContext) {
		Log.d(TAG, "Registering callback");
		cordova.getThreadPool().execute(new Runnable() {
			public void run() {
				Log.d(TAG, "Registering Read Callback");
				readCallback = callbackContext;
				JSONObject returnObj = new JSONObject();
				addProperty(returnObj, "registerReadCallback", "true");
				// Keep the callback
				PluginResult pluginResult = new PluginResult(PluginResult.Status.OK, returnObj);
				pluginResult.setKeepCallback(true);
				callbackContext.sendPluginResult(pluginResult);
			}
		});
	}

	/**
	 * Paused activity handler
	 * 
	 * @see org.apache.cordova.CordovaPlugin#onPause(boolean)
	 */
	@Override
	public void onPause(boolean multitasking) {
		if (sleepOnPause) {
			stopIoManager();
			if (portA != null) {
				try {
					portA.setDTR(false);
					portA.setRTS(false);
					portA.close();
				} catch (IOException e) {
					// Ignore
				}

			}
			if (portB != null) {
				try {
					portB.setDTR(false);
					portB.setRTS(false);
					portB.close();
				} catch (IOException e) {
					// Ignore
				}
				// port = null;
			}
		}
	}

	/**
	 * Resumed activity handler
	 * 
	 * @see org.apache.cordova.CordovaPlugin#onResume(boolean)
	 */
	@Override
	public void onResume(boolean multitasking) {
		Log.d(TAG, "Resumed, driver=" + driver);
		if (sleepOnPause) {
			if (driver == null) {
				Log.d(TAG, "No serial device to resume.");
			} else {
				UsbDeviceConnection connection = manager.openDevice(driver.getDevice());
				if (connection != null) {
					// get first port and open it
					portA = driver.getPorts().get(0);
					try {
						portA.open(connection);
						portA.setParameters(baudRateA, dataBitsA, stopBitsA, parityA);
						if (setDTRA)
							portA.setDTR(true);
						if (setRTSA)
							portA.setRTS(true);
					} catch (IOException e) {
						// deal with error
						Log.d(TAG, e.getMessage());
					}
					Log.d(TAG, "Serial port A opened!");

					portB = driver.getPorts().get(1);
					try {
						portB.open(connection);
						portB.setParameters(baudRateB, dataBitsB, stopBitsB, parityB);
						if (setDTRB)
							portB.setDTR(true);
						if (setRTSB)
							portB.setRTS(true);
					} catch (IOException e) {
						// deal with error
						Log.d(TAG, e.getMessage());
					}
					Log.d(TAG, "Serial port B opened!");
				} else {
					Log.d(TAG, "Cannot connect to the device!");
				}
				Log.d(TAG, "Serial device: " + driver.getClass().getSimpleName());
			}

			onDeviceStateChange();
		}
	}

	/**
	 * Destroy activity handler
	 * 
	 * @see org.apache.cordova.CordovaPlugin#onDestroy()
	 */
	@Override
	public void onDestroy() {
		Log.d(TAG, "Destroy, port=A" + portA);
		if (portA != null) {
			try {
				portA.setDTR(false);
				portA.setRTS(false);
				portA.close();
			} catch (IOException e) {
				Log.d(TAG, e.getMessage());
			}
		}
		Log.d(TAG, "Destroy, port=B" + portB);
		if (portB != null) {
			try {
				portB.setDTR(false);
				portB.setRTS(false);
				portB.close();
			} catch (IOException e) {
				Log.d(TAG, e.getMessage());
			}
		}
		onDeviceStateChange();
	}

	/**
	 * Utility method to add some properties to a {@link JSONObject}
	 * 
	 * @param obj   the json object where to add the new property
	 * @param key   property key
	 * @param value value of the property
	 */
	private void addProperty(JSONObject obj, String key, Object value) {
		try {
			obj.put(key, value);
		} catch (JSONException e) {
		}
	}

	/**
	 * Utility method to add some properties to a {@link JSONObject}
	 * 
	 * @param obj   the json object where to add the new property
	 * @param key   property key
	 * @param bytes the array of byte to add as value to the {@link JSONObject}
	 */
	private void addPropertyBytes(JSONObject obj, String key, byte[] bytes) {
		String string = Base64.encodeToString(bytes, Base64.NO_WRAP);
		this.addProperty(obj, key, string);
	}
}