import { useMemo } from 'react';
import { useTable, useSortBy } from 'react-table';
import moment from 'moment';
import './Table.css';

const Table = ({ data, onRowButtonClick }) => {
  const columns = useMemo(() => {
    // Если данных нет, возвращаем пустой массив колонок
    if (data.length === 0) return [];

    // Получаем первый элемент данных для определения ключей объекта
    const headerMap = {
      litera: 'СПС Литера',
      spsNumber: 'Заводской номер СПС',
      spsSection: 'Номер секции СПС',
      spsSeriesID: 'Серия СПС',
      uuid: 'ID',
      lastPacketAt: 'Дата последней передачи',
    };

    // Получаем первый элемент данных для определения ключей объекта
    const longestObject = data.reduce((prev, current) => {
      return Object.keys(current).length > Object.keys(prev).length ? current : prev;
    }, {});

    // Маппинг ключей объекта на конфигурацию колонок
    const columnsConfig = Object.keys(longestObject).map((key) => ({
      Header: headerMap[key] || key,
      accessor: key,
      Cell: ({ value }) => {
        // Проверка, если ключ равен 'dt',
        if ((key === 'dt' || key === 'lastPacketAt') && typeof value === 'string') {
          return moment(value).format('DD.MM.YYYY HH:mm:ss');
        }
        // Проверка, является ли значение объектом (и не датой, так как даты тоже являются объектами)
        if (value && typeof value === 'object' && !(value instanceof Date)) {
          return JSON.stringify(value, null, 2); // Преобразование объекта в строку JSON
        }
        return value;
      },
    }));

    // Добавление кликабельной ячейки, если есть обработчик клика
    if (onRowButtonClick) {
      // Если передана функция onRowButtonClick, делаем первую ячейку строки кликабельной
      columnsConfig[0].Cell = ({ row }) => (
        <div className="clickable-cell" onClick={() => onRowButtonClick(row.original)}>
          {row.original[columnsConfig[0].accessor]}
        </div>
      );
    }

    return columnsConfig;
  }, [data, onRowButtonClick]);
  // Хуки React Table для управления таблицей
  const { getTableProps, getTableBodyProps, headerGroups, rows, prepareRow } = useTable({ columns, data }, useSortBy);

  // Отображение сообщения, если данных нет
  if (data.length === 0) {
    return (
      <div className="center">
        <p>Данные отсутствуют</p>
      </div>
    );
  }

  return (
    <table {...getTableProps()} className="styled-table">
      <thead>
        {headerGroups.map((headerGroup) => (
          <tr {...headerGroup.getHeaderGroupProps()}>
            {headerGroup.headers.map((column) => (
              <th {...column.getHeaderProps(column.getSortByToggleProps())}>
                {column.render('Header')}
                <span>{column.isSorted ? (column.isSortedDesc ? ' ⬇️' : ' ⬆️') : ''}</span>
              </th>
            ))}
          </tr>
        ))}
      </thead>
      <tbody {...getTableBodyProps()}>
        {rows.map((row) => {
          prepareRow(row);
          return (
            <tr {...row.getRowProps()}>
              {row.cells.map((cell) => {
                return <td {...cell.getCellProps()}>{cell.render('Cell')}</td>;
              })}
            </tr>
          );
        })}
      </tbody>
    </table>
  );
};

export default Table;
