const Devices = require('../db/models/devices');
const logger = require('../utils/logger');

async function initDevicesFromServer() {
  try {
    const fieldsToSelect = '-__v -_id -isAuthorized -createdAt -updatedAt -lastPacketAt';
    const allDevices = await Devices.find({ isAuthorized: true }).select(fieldsToSelect);
    const forEditDevices = await Devices.find({ isAuthorized: false }).select(fieldsToSelect);
    return { allDevices, forEditDevices };
  } catch (error) {
    logger.error('Error during retrieving devices from server:', error);
    throw error;
  }
}

async function createDeviceIfNotExists(uuid) {
  try {
    const existingDevice = await Devices.findOne({ uuid });
    if (!existingDevice) {
      const newDevice = new Devices({ uuid });
      await newDevice.save();
      logger.info(`New device created with uuid: ${uuid}`);
    }
  } catch (error) {
    logger.error('Error during checking/creating device:', error);
    throw error;
  }
}

async function editDevicesFromServer(payload) {
  try {
    const updatedDevice = await Devices.findOneAndUpdate(
      { uuid: payload.uuid },
      {
        spsSeriesID: payload.spsSeriesID,
        spsNumber: payload.spsNumber,
        spsSection: payload.spsSection,
        litera: payload.litera,
        isAuthorized: payload.isAuthorized,
      },
      { new: true, runValidators: true, select: '-__v -_id -isAuthorized -createdAt -updatedAt -lastPacketAt' },
    );
    return updatedDevice;
  } catch (error) {
    logger.error('Error during saving devices to server:', error);
    throw error;
  }
}

async function checkLastPacketReceivedFromDb() {
  const thirtyMinutesAgo = new Date(Date.now() - 30 * 60 * 1000);
  const inactiveDevices = await Devices.find({
    lastPacketAt: { $lt: thirtyMinutesAgo },
    isAuthorized: true,
  }).select('-__v -_id -isAuthorized -createdAt -updatedAt');
  return inactiveDevices;
}

module.exports = {
  initDevicesFromServer,
  editDevicesFromServer,
  createDeviceIfNotExists,
  checkLastPacketReceivedFromDb,
};
